#!/usr/bin/env python
#
#    backup.py v1.0.1
#
#    Copyright (C) 2006  Michele <o-zone@zerozone.it> Pinassi
#
#    This software is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This software is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# 24-Feb-06 Added patch by k0wax <k0wax@mail.ru>

import os, sys, string,time
from stat import *
import shutil
import smtplib

class BackupClass:

# === HARDCODED STUFF
    # where to store temporary files (this dir will be created when backup start and removed whn backup finish)
    tmpDir = "/tmp/backup"
    # where i can found main.cfg
    cfgDir = "/var/spool/backup"
    # what SMTP server i have to use for notify
    SMTPServer = "localhost"

# ========================= DO NOT CHANGE ANYTHING BELOW THIS LINE ============================

    sumFileName = "/summary.log"

    backupDirs = []
    fileDB = []
    excludeExt = []

    backupDir = ""
    backupName = "backup-"+time.strftime("%d%m%Y%H%M") # Default backup name
    backupMode = 0
    notifyMail = []
    notifyFrom = ""

    filesCopied = 0
    filesChecked = 0
    filesSkipped = 0
    filesRemoved = 0
    filesAdded = 0
    totalDirectories = 0

    def loadConfig(self):
	try:
	    cfgFile = open(self.cfgDir+"/main.cfg")
	except Exception,e:
	    print "Error %s"%e
	    return False
	line = string.strip(cfgFile.readline())
        while line != "":
    	    if(line[0] != "#"):
		try:
		    (com,arg) = string.split(line,' ',1)
	        except Exception,e:
	    	    print "Error %s"%e
		else:
		    if(com == "TARGETDIR"):
			self.backupDirs.append(arg)
	    	    if(com == "EXCLUDEEXT"):
			self.excludeExt.append(arg)
	    	    if(com == "NOTIFYMAIL"):
			self.notifyMail.append(arg)
		    if(com == "BACKUPDIR"):
			self.backupDir = arg
		    if(com == "BACKUPNAME"):
			self.backupName = time.strftime(arg)
		    if(com == "NOTIFYFROM"):
			self.notifyFrom = arg
			
	    line = string.strip(cfgFile.readline())
	cfgFile.close()
	return True
    
    def loadFileDB(self):
	try:
	    cfgFile = open(self.cfgDir+"/file.db")
	except Exception,e:
	    print "Error %s"%e
	    return False
	line = string.strip(cfgFile.readline())
	while line != "":
	    if(line[0] != "#"):
		try:
		    file = string.split(line,' ',2)
	        except Exception,e:
	    	    print "Error %s"%e
	        else:
		    file.append(0)
	    	    self.fileDB.append(file)
	    line = string.strip(cfgFile.readline())
	cfgFile.close()
	return True

    def saveFileDB(self):
	try:
	    cfgFile = open(self.cfgDir+"/file.db","w")
	except Exception,e:
	    print "Error %s"%e
	    return
	for dbEntry in self.fileDB:
	    cfgFile.write("%s %s %s\n"%(dbEntry[0],dbEntry[1],dbEntry[2]))
	cfgFile.close()

    def fileIsChanged(self,filePath,fileStat):
	if self.backupMode == 1: # Se backup completo...
	    return True
	fileSize = fileStat[ST_SIZE]
	fileChanged = fileStat[ST_MTIME]
	for dbEntry in self.fileDB:
	    if(dbEntry[0] == filePath):
		# Analizza il file se  cambiato
	        if(fileSize != int(dbEntry[1])):
		    return True
	        if(fileChanged != int(dbEntry[2])):
		    return True
		dbEntry[3] = 1
		return False # Ok, non copiare
	return True
    
    def updateFileEntry(self,filePath,fileStat):
	fileSize = fileStat[ST_SIZE]
	fileChanged = fileStat[ST_MTIME]
	for dbEntry in self.fileDB:
	    if(dbEntry[0] == filePath):
	    	dbEntry[1] = fileSize
		dbEntry[2] = fileChanged
		dbEntry[3] = 2
		return
	self.fileDB.append((filePath,fileSize,fileChanged,3))
        
    def cbFile(self,filePath,fileStat):
	self.filesChecked = self.filesChecked + 1
	# Controlla che il file sia abilitato
	fileExt = os.path.splitext(filePath)[1]
	if fileExt not in self.excludeExt:
	    if self.fileIsChanged(filePath,fileStat):
		if self.copyFile(filePath):
		    # Se copia andata a buon fine, aggiorna nel DB
		    self.updateFileEntry(filePath,fileStat)
	else:
	    self.filesSkipped = self.filesSkipped + 1

    def copyFile(self,filePath):
	print "Copying %s..."%filePath,
	try:
	    shutil.copy2(filePath, self.tmpDir+filePath)
	except Exception,e:
	    print "ERROR(%s)"%e
	    return False
	self.filesCopied = self.filesCopied + 1
	print "OK"
	return True

    def walktree(self,top, callback):
	self.totalDirectories = self.totalDirectories + 1
	# Crea la directory nella struttura temporanea
        try:
    	    os.makedirs(self.tmpDir+top)
        except Exception,e:
    	    print "Directory not created: %s"%e
	# Analizza la lista dei files
	for f in os.listdir(top):
	    pathname = os.path.join(top, f)
	    try:
		fileStat = os.stat(pathname)
	    except Exception,e:
		print "Skipping infalid file %s (%s)" % (pathname,e)
		continue		
	    mode = fileStat[ST_MODE]
	    if S_ISDIR(mode):
		# It's a directory, recurse into it
		if os.path.islink(pathname):
		    print "Skipping link %s" % pathname
		else:
        	    self.walktree(pathname, callback)
	    elif S_ISREG(mode):
        	# It's a file, call the callback function
		callback(pathname,fileStat)
	    else:
		# Unknown file type, print a message
		print 'Skipping %s' % pathname

    def __init__(self):
	if self.loadConfig() is not True:
	    print "FATAL: No config file found (%s missing ?)"%self.cfgDir+"/main.cfg"
	    sys.exit(-1)
	
	if self.loadFileDB() is not True:
	    print "WARNING: No FileDB found. Maybe first time run ? Switching to COMPLETE backup mode"
	    self.backupMode = 1
    
    def sendReport(self):
	msg = "From: %s\r\nTo: %s\r\nSubject: %s\r\n\r\n"%(self.notifyFrom, ", ".join(self.notifyMail),"Backup summary "+time.asctime())
	if self.backupMode == 1:
	    msg = msg + "COMPLETE Backup summary\r\n\r\n"
	else:
	    msg = msg + "INCREMENTAL Backup summary\r\n\r\n"

	msg = msg + "Total files: %d\tTotal Directories: %d\r\n"%(self.filesChecked,self.totalDirectories)
	msg = msg + "Changed files: %d\tAdded:%d\tRemoved:%d\tSkipped: %d\r\n"%(self.filesCopied,self.filesAdded,self.filesRemoved,self.filesSkipped)
	
	msg = msg + "\r\nHave a nice day.\r\n"
    
	try:
	    server = smtplib.SMTP(self.SMTPServer)
	    server.set_debuglevel(False)
	    server.sendmail(self.notifyFrom, self.notifyMail, msg)
	    server.quit()
	except Exception,e:
	    print "SEND NOTIFY Error: %s"%e

    def startBackup(self):
	# Avvia analisi delle directory di cui fare il backup
	for bckDir in self.backupDirs:
	    self.walktree(bckDir,self.cbFile)

	# Apri summary file
	try:
	    sumFile = open(self.tmpDir+self.sumFileName,"w")
	except Exception,e:
	    print "Error %s"%e
	    return

	for dbEntry in self.fileDB:
	    if dbEntry[3] == 0: # DELETED
		print "File %s deleted."%dbEntry[0]
		self.filesRemoved = self.filesRemoved + 1
		sumFile.write("REM %s\n"%dbEntry[0])
	    if dbEntry[3] == 1: # SAME
		pass	    
	    if dbEntry[3] == 2: # CHANGED
		sumFile.write("CHG %s\n"%dbEntry[0])
	    if dbEntry[3] == 3: # ADDED
		self.filesAdded = self.filesAdded + 1
		print "File %s added."%dbEntry[0]
		sumFile.write("ADD %s\n"%dbEntry[0])
    
	sumFile.close()
    
	if self.filesCopied > 0:
	    if self.backupMode == 1:
		backupTGZ = os.path.join(self.backupDir, self.backupName + ".tar.gz");
	    else:
		backupTGZ = os.path.join(self.backupDir, self.backupName + ".diff.tar.gz");
	    # Comprimi e crea l'archivio
	    os.spawnlp(os.P_WAIT, 'tar','tar','czf',backupTGZ,self.tmpDir)

    
	print "======================================="
	if self.backupMode == 1:
	    print "COMPLETE Backup summary"
	else:
	    print "INCREMENTAL Backup summary"

	print "Total files: %d\tTotal Directories: %d"%(self.filesChecked,self.totalDirectories)
	print "Changed files: %d\tAdded:%d\tRemoved:%d\tSkipped: %d"%(self.filesCopied,self.filesAdded,self.filesRemoved,self.filesSkipped)
	# Salva DB dei files
	self.saveFileDB()
	# Rimuovi directory temporanea
	try:
	    shutil.rmtree(self.tmpDir)
	except Exception,e:
	    print "Exception: %s"%e
	# And finally send report by mail
	self.sendReport()

if __name__=='__main__':
    backup = BackupClass()
    backup.startBackup()
    